@@ -45,7 +45,6 @@
 #include <xercesc/util/XMLUniDefs.hpp>
 #include <xercesc/util/regx/RegularExpression.hpp>
 
-using shibspconstants::SHIB2SPCONFIG_NS;
 using namespace shibsp;
 using namespace xmltooling;
 using namespace boost;
@@ -76,12 +75,7 @@ namespace shibsp {
         ~Override() {}
 
         // Provides filter to exclude special config elements.
-#ifdef SHIBSP_XERCESC_SHORT_ACCEPTNODE
-        short
-#else
-        FilterAction
-#endif
-        acceptNode(const DOMNode* node) const {
+        FilterAction acceptNode(const DOMNode* node) const {
             return FILTER_REJECT;
         }
 
@@ -128,7 +122,8 @@ namespace shibsp {
     class XMLRequestMapper : public RequestMapper, public ReloadableXMLFile
     {
     public:
-        XMLRequestMapper(const DOMElement* e) : ReloadableXMLFile(e,Category::getInstance(SHIBSP_LOGCAT ".RequestMapper")) {
+        XMLRequestMapper(const DOMElement* e, bool deprecationSupport=true)
+            : ReloadableXMLFile(e, Category::getInstance(SHIBSP_LOGCAT ".RequestMapper"), true, deprecationSupport) {
             background_load();
         }
 
@@ -149,9 +144,9 @@ namespace shibsp {
     #pragma warning( pop )
 #endif
 
-    RequestMapper* SHIBSP_DLLLOCAL XMLRequestMapperFactory(const DOMElement* const & e)
+    RequestMapper* SHIBSP_DLLLOCAL XMLRequestMapperFactory(const DOMElement* const & e, bool deprecationSupport)
     {
-        return new XMLRequestMapper(e);
+        return new XMLRequestMapper(e, deprecationSupport);
     }
 
     static const XMLCh _AccessControl[] =           UNICODE_LITERAL_13(A,c,c,e,s,s,C,o,n,t,r,o,l);
@@ -186,17 +181,18 @@ RequestMapper::~RequestMapper()
 
 void Override::loadACL(const DOMElement* e, Category& log)
 {
+    bool deprecationSupport = e ? XMLString::equals(e->getNamespaceURI(), shibspconstants::SHIB2SPCONFIG_NS) : false;
     try {
         const DOMElement* acl = XMLHelper::getFirstChildElement(e,htaccess);
         if (acl) {
             log.info("building Apache htaccess AccessControl provider...");
-            m_acl.reset(SPConfig::getConfig().AccessControlManager.newPlugin(HT_ACCESS_CONTROL,acl));
+            m_acl.reset(SPConfig::getConfig().AccessControlManager.newPlugin(HT_ACCESS_CONTROL,acl, deprecationSupport));
         }
         else {
             acl = XMLHelper::getFirstChildElement(e,_AccessControl);
             if (acl) {
                 log.info("building XML-based AccessControl provider...");
-                m_acl.reset(SPConfig::getConfig().AccessControlManager.newPlugin(XML_ACCESS_CONTROL,acl));
+                m_acl.reset(SPConfig::getConfig().AccessControlManager.newPlugin(XML_ACCESS_CONTROL,acl, deprecationSupport));
             }
             else {
                 acl = XMLHelper::getFirstChildElement(e,AccessControlProvider);
@@ -204,7 +200,7 @@ void Override::loadACL(const DOMElement* e, Category& log)
                     string t(XMLHelper::getAttrString(acl, nullptr, _type));
                     if (!t.empty()) {
                         log.info("building AccessControl provider of type %s...", t.c_str());
-                        m_acl.reset(SPConfig::getConfig().AccessControlManager.newPlugin(t.c_str(), acl));
+                        m_acl.reset(SPConfig::getConfig().AccessControlManager.newPlugin(t.c_str(), acl, deprecationSupport));
                     }
                     else {
                         throw ConfigurationException("<AccessControlProvider> missing type attribute.");
@@ -213,7 +209,7 @@ void Override::loadACL(const DOMElement* e, Category& log)
             }
         }
     }
-    catch (std::exception& ex) {
+    catch (const std::exception& ex) {
         log.crit("exception building AccessControl provider: %s", ex.what());
         m_acl.reset(new AccessControlDummy());
     }
@@ -223,7 +219,8 @@ Override::Override(bool unicodeAware, const DOMElement* e, Category& log, const
     : m_unicodeAware(unicodeAware)
 {
     // Load the property set.
-    load(e, nullptr, this);
+    xmltooling::QName unsetter(nullptr, "unset");
+    load(e, nullptr, this, nullptr, &unsetter);
     setParent(base);
 
     // Load any AccessControl provider.
@@ -261,7 +258,7 @@ Override::Override(bool unicodeAware, const DOMElement* e, Category& log, const
 
             if (*n) {
                 // Create a placeholder Path element for the first path segment and replant under it.
-                DOMElement* newpath = path->getOwnerDocument()->createElementNS(shibspconstants::SHIB2SPCONFIG_NS, Path);
+                DOMElement* newpath = path->getOwnerDocument()->createElementNS(path->getNamespaceURI(), Path);
                 newpath->setAttributeNS(nullptr, name, namebuf.c_str());
                 path->setAttributeNS(nullptr, name, n);
                 path->getParentNode()->replaceChild(newpath, path);
@@ -298,7 +295,7 @@ Override::Override(bool unicodeAware, const DOMElement* e, Category& log, const
             }
             free(dup);
         }
-        catch (std::exception&) {
+        catch (const std::exception&) {
             free(dup);
             throw;
         }
@@ -320,7 +317,7 @@ Override::Override(bool unicodeAware, const DOMElement* e, Category& log, const
             if (path && path->hasAttributeNS(nullptr, ignoreCase)) {
                 // In this one case, we've left ignoreCase reversed (true means case sensitive, false means insensitive).
                 // This was to protect people who followed the security advisory for SSPCPP-691 and reversed their setting.
-                log.error("Deprecated ignoreCase attribute in PathRegex element will be interpreted backwards. Replace with caseSensitive");
+                log.error("DEPRECATED: ignoreCase attribute in PathRegex element will be interpreted backwards. Replace with caseSensitive");
                 caseSensitive = XMLHelper::getAttrBool(path, true, ignoreCase);
             } else {
                 // If the old ignoreCase setting isn't set, then we just process normally.
@@ -330,7 +327,7 @@ Override::Override(bool unicodeAware, const DOMElement* e, Category& log, const
                 boost::shared_ptr<RegularExpression> re(new RegularExpression(n, caseSensitive ? &chNull : caseInsensitiveOption));
                 m_regexps.push_back(make_pair(re, o));
             }
-            catch (XMLException& ex) {
+            catch (const XMLException& ex) {
                 auto_ptr_char tmp(ex.getMessage());
                 log.error("caught exception while parsing PathRegex regular expression (%d): %s", i, tmp.get());
                 throw ConfigurationException("Invalid regular expression in PathRegex element.");
@@ -357,7 +354,7 @@ Override::Override(bool unicodeAware, const DOMElement* e, Category& log, const
             boost::shared_ptr<RegularExpression> re((v && *v) ? new RegularExpression(v) : nullptr);
             m_queries.push_back(boost::make_tuple(string(ntemp.get()), re, o));
         }
-        catch (XMLException& ex) {
+        catch (const XMLException& ex) {
             auto_ptr_char tmp(ex.getMessage());
             log.error("caught exception while parsing Query regular expression (%d): %s", i, tmp.get());
             throw ConfigurationException("Invalid regular expression in Query element.");
@@ -367,6 +364,16 @@ Override::Override(bool unicodeAware, const DOMElement* e, Category& log, const
     }
 }
 
+static char _x2c(const char *what)
+{
+    register char digit;
+
+    digit = (what[0] >= 'A' ? ((what[0] & 0xdf) - 'A')+10 : (what[0] - '0'));
+    digit *= 16;
+    digit += (what[1] >= 'A' ? ((what[1] & 0xdf) - 'A')+10 : (what[1] - '0'));
+    return(digit);
+}
+
 const Override* Override::locate(const HTTPRequest& request) const
 {
     // This function is confusing because it's *not* recursive.
@@ -374,25 +381,55 @@ const Override* Override::locate(const HTTPRequest& request) const
     // path parameter starts with the entire request path and
     // we can skip the leading slash as irrelevant.
     const char* path = request.getRequestURI();
-    if (*path == '/')
+    if (path && *path == '/')
         path++;
 
+    // Fix for bug 574, secadv 20061002
+    // Unescape URI up to query string delimiter by looking for %XX escapes.
+    // Adapted from Apache's util.c, ap_unescape_url function.
+    string dup;
+    if (path) {
+        while (*path) {
+            if (*path == '?') {
+                dup += path;
+                break;
+            }
+            else if (*path != '%') {
+                dup += *path;
+            }
+            else {
+                ++path;
+                if (!isxdigit(*path) || !isxdigit(*(path+1)))
+                    throw ConfigurationException("Bad request URI, contained unsupported encoded characters.");
+                dup += _x2c(path);
+                ++path;
+            }
+            ++path;
+        }
+    }
+
     // Now we copy the path, chop the query string, and possibly lower case it.
-    string dup(path);
     string::size_type sep = dup.find('?');
-    if (sep != string::npos)
+    if (sep != string::npos) {
         dup = dup.substr(0, sep);
-    if (!m_unicodeAware) {
-        to_lower(dup);
     }
 
     // Default is for the current object to provide settings.
     const Override* o = this;
 
+    // Reset the path pointer to the beginning of the decoded copy.
+    path = dup.c_str();
+
     // Tokenize the path by segment and try and map each segment.
     tokenizer< char_separator<char> > tokens(dup, char_separator<char>("/"));
     for (tokenizer< char_separator<char> >::iterator token = tokens.begin(); token != tokens.end(); ++token) {
-        map< string,boost::shared_ptr<Override> >::const_iterator i = o->m_map.find(*token);
+
+        string tokendup(*token);
+        if (!m_unicodeAware) {
+            to_lower(tokendup);
+        }
+
+        map< string,boost::shared_ptr<Override> >::const_iterator i = o->m_map.find(tokendup);
         if (i == o->m_map.end())
             break;  // Once there's no match, we've consumed as much of the path as possible here.
         // We found a match, so reset the settings pointer.
@@ -400,25 +437,20 @@ const Override* Override::locate(const HTTPRequest& request) const
 
         // We descended a step down the path, so we need to advance the original
         // parameter for the regex step later.
-        path += token->length();
+        path += tokendup.length();
         if (*path == '/')
             path++;
     }
 
     // If there's anything left, we try for a regex match on the rest of the path minus the query string.
     if (*path) {
-        string path2(path);
-        sep = path2.find('?');
-        if (sep != string::npos)
-            path2 = path2.substr(0, sep);
-
         for (vector< pair< boost::shared_ptr<RegularExpression>,boost::shared_ptr<Override> > >::const_iterator re = o->m_regexps.begin(); re != o->m_regexps.end(); ++re) {
             try {
-                if (re->first->matches(path2.c_str())) {
+                if (re->first->matches(path)) {
                     o = re->second.get();
                     break;
                 }
-            } catch (XMLException& ex) {
+            } catch (const XMLException& ex) {
                 auto_ptr_char tmp(ex.getMessage());
                 throw ConfigurationException("Caught exception while matching PathRegex : $1", params(1, tmp.get()));
             }
@@ -471,11 +503,18 @@ XMLRequestMapperImpl::XMLRequestMapperImpl(const DOMElement* e, Category& log) :
 #endif
     static const XMLCh _RequestMap[] =  UNICODE_LITERAL_10(R,e,q,u,e,s,t,M,a,p);
 
-    if (e && !XMLHelper::isNodeNamed(e, SHIB2SPCONFIG_NS, _RequestMap))
+    if (e && !XMLHelper::isNodeNamed(e, shibspconstants::SHIB2SPCONFIG_NS, _RequestMap)
+          && !XMLHelper::isNodeNamed(e, shibspconstants::SHIB3SPCONFIG_NS, _RequestMap)) {
         throw ConfigurationException("XML RequestMapper requires conf:RequestMap at root of configuration.");
+    }
+
+    if (XMLString::equals(e->getNamespaceURI(), shibspconstants::SHIB2SPCONFIG_NS)) {
+        log.warn("DEPRECATED: legacy 2.0 configuration, support will be removed from a future version of the software");
+    }
 
     // Load the property set.
-    load(e, nullptr, this);
+    xmltooling::QName unsetter(nullptr, "unset");
+    load(e, nullptr, this, nullptr, &unsetter);
 
     // Inject "default" app ID if not explicit.
     if (!getString("applicationId").first)
@@ -505,7 +544,7 @@ XMLRequestMapperImpl::XMLRequestMapperImpl(const DOMElement* e, Category& log) :
                 );
             m_regexps.push_back(make_pair(re, o));
         }
-        catch (XMLException& ex) {
+        catch (const XMLException& ex) {
             auto_ptr_char tmp(ex.getMessage());
             log.error("caught exception while parsing HostRegex regular expression (%d): %s", i, tmp.get());
         }
@@ -674,7 +713,7 @@ RequestMapper::Settings XMLRequestMapper::getSettings(const HTTPRequest& request
         const Override* o = m_impl->findOverride(vhost.c_str(), request);
         return Settings(o, o->getAC());
     }
-    catch (XMLException& ex) {
+    catch (const XMLException& ex) {
         auto_ptr_char tmp(ex.getMessage());
         m_log.error("caught exception while locating content settings: %s", tmp.get());
         throw ConfigurationException("XML-based RequestMapper failed to retrieve content settings.");
